//------------------------------------------------------------------------------
// SimRobot.cs
//
// Ben Axelrod  -  08/2007
//
//------------------------------------------------------------------------------
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Xml;
using submgr = Microsoft.Dss.Services.SubscriptionManager;

#region Simulation namespaces
using Microsoft.Robotics.Simulation;
using Microsoft.Robotics.Simulation.Engine;
using engineproxy = Microsoft.Robotics.Simulation.Engine.Proxy;
using Microsoft.Robotics.Simulation.Physics;

using Microsoft.Robotics.PhysicalModel;
using System.ComponentModel;
using Microsoft.Dss.Core.DsspHttp;
using System.Net;
using W3C.Soap;
#endregion


using drive = Microsoft.Robotics.Services.Simulation.Drive.Proxy;
using simlrf = Microsoft.Robotics.Services.Simulation.Sensors.LaserRangeFinder.Proxy;
using sicklrf = Microsoft.Robotics.Services.Sensors.SickLRF.Proxy;
using bumper = Microsoft.Robotics.Services.Simulation.Sensors.Bumper.Proxy;
using simwebcam = Microsoft.Robotics.Services.Simulation.Sensors.SimulatedWebcam.Proxy;

using position = Microsoft.Robotics.Services.SimplePosition.Proxy;


namespace Robotics.SimRobot
{
    
    /// <summary>
    /// The Sim Robot Service
    /// </summary>
    [DisplayName("SimRobot")]
    [Description("The SimRobot Service")]
    [Contract(Contract.Identifier)]
    [AlternateContract(position.Contract.Identifier)]
    public class SimRobotService : DsspServiceBase
    {
        /// <summary>
        /// _state
        /// </summary>
        private SimRobotState _state = new SimRobotState();

        /// <summary>
        /// _main Port
        /// </summary>
        [ServicePort("/simrobot", AllowMultipleInstances=false)]
        private SimRobotOperations _mainPort = new SimRobotOperations();

        //alternate service port
        [AlternateServicePort("/2dposition", AllowMultipleInstances = false, AlternateContract = position.Contract.Identifier)]
        private position.SimplePositionOperations _positionPort = new position.SimplePositionOperations();

        //alternate state
        private position.SimplePositionState _positionState = new position.SimplePositionState();
        
        // partner attribute will cause simulation engine service to start
        [Partner("Engine",
            Contract = engineproxy.Contract.Identifier,
            CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate)]
        private engineproxy.SimulationEnginePort _engineServicePort =
            new engineproxy.SimulationEnginePort();

        [Partner("mySubMgr", Contract = submgr.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.CreateAlways)]
        submgr.SubscriptionManagerPort _subMgrPort = new submgr.SubscriptionManagerPort();

        [Partner("positionSubMgr", Contract = submgr.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.CreateAlways)]
        submgr.SubscriptionManagerPort _positionSubMgrPort = new submgr.SubscriptionManagerPort();


        private Pioneer3DX robotBaseEntity;
        private int boxes = 0;
        private static readonly int poseUpdateHz = 15;

        private Port<DateTime> _timerPort = new Port<DateTime>();

        /// <summary>
        /// Default Service Constructor
        /// </summary>
        public SimRobotService(DsspServiceCreationPort creationPort) : 
                base(creationPort)
        {
        }

        /// <summary>
        /// Service Start
        /// </summary>
        protected override void Start()
        {
            #region base.Start()
            // Listen on the main port for requests and call the appropriate handler.
            Interleave mainInterleave = ActivateDsspOperationHandlers();

            // Publish the service to the local Node Directory
            DirectoryInsert();

            // display HTTP service Uri
            LogInfo(LogGroups.Console, "Service uri: ");
            #endregion

            //initialize _state

            // orient sim camera view point
            SetupCamera();
            
            // Add objects (entities) in our simulated world
            PopulateWorld();


            //alternate service port
            //Activate(
            //    new Interleave(
            //        new ExclusiveReceiverGroup(
            //            Arbiter.ReceiveWithIterator<position.Replace>(true, _positionPort, PositionReplaceHandler)
            //        ),
            //        new ConcurrentReceiverGroup(
            //            Arbiter.ReceiveWithIterator<position.Get>(true, _positionPort, PositionGetHandler),
            //            Arbiter.ReceiveWithIterator<position.Subscribe>(true, _positionPort, PositionSubscribeHandler),
            //            Arbiter.ReceiveWithIterator<position.ReliableSubscribe>(true, _positionPort, PositionReliableSubscribeHandler)
            //        )
            //    )
            //);

            
            //start monitoring the robot's global pose
            _timerPort.Post(DateTime.Now);
            Activate(Arbiter.Receive(true, _timerPort, TimerHandler));
        }

        #region Standard Operation Handlers

       
        /// <summary>
        /// Get Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> GetHandler(Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReplaceHandler(Replace replace)
        {
            _state = replace.Body;
            replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);

            // issue notification
            base.SendNotification(_subMgrPort, replace);

            yield break;
        }

        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> SubscribeHandler(Subscribe subscribe)
        {
            SubscribeRequestType request = subscribe.Body;
            LogInfo("Subscribe request from: " + request.Subscriber);

            yield return Arbiter.Choice(
                SubscribeHelper(_subMgrPort, request, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    base.SendNotification<Replace>(_subMgrPort, request.Subscriber, _state);
                },
                delegate(Exception e)
                {
                    LogError(null, "Subscribe failed", e);
                }
            );

            yield break;
        }

        #endregion


        void TimerHandler(DateTime signal)
        {
            //make a new state
            position.SimplePositionState newstate = new position.SimplePositionState();
            newstate.TimeStamp = DateTime.Now;
            newstate.X = robotBaseEntity.State.Pose.Position.X;
            newstate.Y = -robotBaseEntity.State.Pose.Position.Z;
            newstate.Theta = ConvertQuaternion(robotBaseEntity.State.Pose.Orientation);

            //send a replace message to ourselves with the new state
            _positionPort.Post(new position.Replace(newstate));

            //call ourselves again after small timeout
            Activate(
                Arbiter.Receive(false, TimeoutPort((int)Math.Round(1000.0 / (double)poseUpdateHz)),
                    delegate(DateTime time)
                    {
                        _timerPort.Post(time);
                    }
                )
            );
        }

        /// <summary>
        /// Convert a Quaternion to an angle around the y axis from 0 to 2 Pi
        /// </summary>
        public double ConvertQuaternion(Quaternion q)
        {
            AxisAngle a = new AxisAngle();
            a = Quaternion.ToAxisAngle(q);

            if (float.IsNaN(a.Angle))
                return 0;
            else if (Math.Sign(a.Axis.Y) < 0)
                return 2.0 * Math.PI - a.Angle;

            return a.Angle;
        }


        #region Position Operation Handlers

        /// <summary>
        /// position Get Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent, PortFieldName = "_positionPort")]
        public IEnumerator<ITask> PositionGetHandler(position.Get get)
        {
            Pose currentPose = robotBaseEntity.State.Pose;

            position.SimplePositionState newstate = new position.SimplePositionState();
            newstate.X = currentPose.Position.X;
            newstate.Y = -currentPose.Position.Z;

            newstate.Theta = (float)ConvertQuaternion(currentPose.Orientation);

            newstate.TimeStamp = DateTime.Now;

            get.ResponsePort.Post(newstate); //need to return most up to date state info
            yield break;
        }

        /// <summary>
        /// position Replace handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive, PortFieldName = "_positionPort")]
        public IEnumerator<ITask> PositionReplaceHandler(position.Replace replace)
        {
            _positionState = replace.Body;
            replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);

            // issue notification
            base.SendNotification(_positionSubMgrPort, replace);

            yield break;
        }

        /// <summary>
        /// position Subscribe handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent, PortFieldName = "_positionPort")]
        public IEnumerator<ITask> PositionSubscribeHandler(position.Subscribe subscribe)
        {
            SubscribeRequestType request = subscribe.Body;
            LogInfo("Subscribe request from: " + request.Subscriber);

            yield return Arbiter.Choice(
                SubscribeHelper(_positionSubMgrPort, request, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    base.SendNotification<position.Replace>(_positionSubMgrPort, request.Subscriber, new position.Replace(_positionState));
                },
                delegate(Exception e)
                {
                    subscribe.ResponsePort.Post(new Fault());
                    LogError(null, "Subscribe failed", e);
                }
            );

            yield break;
        }

        /// <summary>
        /// position Reliable Subscribe handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent, PortFieldName = "_positionPort")]
        public IEnumerator<ITask> PositionReliableSubscribeHandler(position.ReliableSubscribe subscribe)
        {
            SubscribeRequestType request = subscribe.Body;
            LogInfo("Subscribe request from: " + request.Subscriber);

            yield return Arbiter.Choice(
                SubscribeHelper(_positionSubMgrPort, request, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    base.SendNotification<position.Replace>(_positionSubMgrPort, request.Subscriber, new position.Replace(_positionState));
                },
                delegate(Exception e)
                {
                    subscribe.ResponsePort.Post(new Fault());
                    LogError(null, "Subscribe failed", e);
                }
            );

            yield break;
        }

        #endregion

        private void SetupCamera()
        {
            // Set up initial view
            CameraView view = new CameraView();
            //view.EyePosition = new Vector3(-1.52f, 4.07f, -2.15f);
            view.EyePosition = new Vector3(5.56f, 12f, 10.5f);
            //view.LookAtPoint = new Vector3(-1.19f, 3.77f, -1.63f);
            view.LookAtPoint = new Vector3(5.56f, 11.1f, 10.1f);
            SimulationEngine.GlobalInstancePort.Update(view);
        }

        private void PopulateWorld()
        {
            AddSky();
            AddGround();

            AddPioneer3DXRobot(new Vector3(1, 0, 1));

            //create room
            AddBox(new Vector3(11.1252f, 1, 0.762f), new Vector2(5.5626f, -0.381f));
            AddBox(new Vector3(0.7112f, 1, 3.683f), new Vector2(-0.381f, 1.0795f));
            AddBox(new Vector3(0.762f, 1, 9.4234f), new Vector2(-1.143f, 3.9497f));
            AddBox(new Vector3(1.4732f, 1, 1.4732f), new Vector2(-0.7874f, 9.4488f));
            AddBox(new Vector3(4.2418f, 1, 0.762f), new Vector2(2.1209f, 9.8044f));
            AddBox(new Vector3(0.762f, 1, 3.683f), new Vector2(11.5062f, 1.0795f));
            AddBox(new Vector3(0.762f, 1, 9.4234f), new Vector2(12.2682f, 3.937f));
            AddBox(new Vector3(1.4732f, 1, 1.4732f), new Vector2(11.8999f, 9.4488f));
            AddBox(new Vector3(1.2192f, 1, 0.762f), new Vector2(10.5156f, 9.8044f));
            AddBox(new Vector3(0.635f, 1, 0.635f), new Vector2(11.5062f, 3.3528f));
            AddBox(new Vector3(0.635f, 1, 0.635f), new Vector2(11.5062f, 4.0259f));
            AddBox(new Vector3(0.635f, 1, 0.635f), new Vector2(11.5062f, 4.699f));
            AddBox(new Vector3(0.635f, 1, 0.635f), new Vector2(11.5062f, 5.3721f));
            AddBox(new Vector3(0.635f, 1, 0.635f), new Vector2(11.5062f, 6.0452f));
            AddBox(new Vector3(0.635f, 1, 0.635f), new Vector2(-0.381f, 3.3528f));
            AddBox(new Vector3(0.635f, 1, 0.635f), new Vector2(-0.381f, 4.0259f));
            AddBox(new Vector3(0.635f, 1, 0.635f), new Vector2(-0.381f, 4.699f));
            AddBox(new Vector3(0.762f, 1, 0.1524f), new Vector2(10.7061f, 0.1143f));

            AddBox(new Vector3(0.5f, 1, 0.5f), new Vector2(5.75f, 9.8f));
            AddBox(new Vector3(0.5f, 1, 0.5f), new Vector2(7f, 9.8f));
            AddBox(new Vector3(0.5f, 1, 0.5f), new Vector2(8.75f, 9.8f));
            AddBox(new Vector3(1, 1, 1), new Vector2(8.25f, 4f));

            AddBox(new Vector3(0.5f, 1, 0.5f), new Vector2(5f, 5f));
            AddBox(new Vector3(0.5f, 1, 0.5f), new Vector2(7f, 9f));
            AddBox(new Vector3(0.5f, 1, 0.5f), new Vector2(8f, 6f));
            AddBox(new Vector3(1, 1, 1), new Vector2(9f, 7f));
            AddBox(new Vector3(0.5f, 1, 0.5f), new Vector2(6f, 8f));

          
            //add extra tables
            AddTable(new Vector3(5f, 0, 0.38f));
            AddTable(new Vector3(6.2f, 0, 0.38f));
            //AddTable(new Vector3(5f, 0, 5f));
            AddTable(new Vector3(5f, 0, 6f));
            AddTable(new Vector3(1.8f, 0, 9f));
        }

        private void AddBox(Vector3 shape, Vector2 position)
        {
            BoxShapeProperties cBoxShape = null;
            SingleShapeEntity cBoxEntity = null;

            cBoxShape = new BoxShapeProperties(10f, new Pose(), shape);
            cBoxShape.Material = new MaterialProperties("gbox", 0.5f, 0.4f, 0.5f);
            cBoxShape.DiffuseColor = new Vector4(0.5f, 0.5f, 0.5f, 1.0f);
            cBoxEntity = new SingleShapeEntity(new BoxShape(cBoxShape), new Vector3(position.X, shape.Y / 2, position.Y));
            cBoxEntity.State.Name = "greybox" + (++boxes);
            SimulationEngine.GlobalInstancePort.Insert(cBoxEntity);
        }

        #region Environment Entities

        void AddSky()
        {
            // Add a sky using a static texture. We will use the sky texture
            // to do per pixel lighting on each simulation visual entity
            SkyEntity sky = new SkyEntity("sky.dds", "sky_diff.dds");
            SimulationEngine.GlobalInstancePort.Insert(sky);

            // Add a directional light to simulate the sun.
            LightEntity sun = new LightEntity();
            sun.Name = "Sun";
            sun.Type = LightEntityType.Directional;
            sun.Color = new Vector4(0.8f, 0.8f, 0.8f, 1);
            sun.Pose.Position = new Vector3(1.0f, 1.0f, -0.5f);
            SimulationEngine.GlobalInstancePort.Insert(sun);
        }

        void AddGround()
        {
            HeightFieldShapeProperties hf = new HeightFieldShapeProperties("height field",
                64, // number of rows 
                100, // distance in meters, between rows
                64, // number of columns
                100, // distance in meters, between columns
                1, // scale factor to multiple height values 
                -1000); // vertical extent of the height field. Should be set to large negative values

            // create array with height samples
            hf.HeightSamples = new HeightFieldSample[hf.RowCount * hf.ColumnCount];
            for (int i = 0; i < hf.RowCount * hf.ColumnCount; i++)
            {
                hf.HeightSamples[i] = new HeightFieldSample();
                hf.HeightSamples[i].Height = (short)(Math.Sin(i * 0.01));
            }

            // create a material for the entire field. We could also specify material per sample.
            hf.Material = new MaterialProperties("ground", 0.8f, 0.5f, 0.8f);

            // insert ground entity in simulation and specify a texture
            SimulationEngine.GlobalInstancePort.Insert(new HeightFieldEntity(hf, "03RamieSc.dds"));
        }

        void AddTable(Vector3 position)
        {
            // create an instance of our custom entity
            TableEntity entity = new TableEntity(position);

            // Name the entity
            entity.State.Name = "table:" + Guid.NewGuid().ToString();

            // Insert entity in simulation. 
            SimulationEngine.GlobalInstancePort.Insert(entity);
        }
        
        #endregion

        #region Pioneer Entity

        void AddPioneer3DXRobot(Vector3 position)
        {
            robotBaseEntity = CreateMotorBase(ref position);

            // Create Laser entity and start simulated laser service
            LaserRangeFinderEntity laser = CreateLaserRangeFinder();
            // insert laser as child to motor base
            robotBaseEntity.InsertEntity(laser);

            // Create bumper array entity and start simulated bumper service
            BumperArrayEntity bumperArray = CreateBumperArray();
            // insert as child of motor base
            robotBaseEntity.InsertEntity(bumperArray);

            // create Camera Entity ans start SimulatedWebcam service
            CameraEntity camera = CreateCamera();
            // insert as child of motor base
            robotBaseEntity.InsertEntity(camera);

            // Finally insert the motor base and its two children 
            // to the simulation
            SimulationEngine.GlobalInstancePort.Insert(robotBaseEntity);

        }

        private Pioneer3DX CreateMotorBase(ref Vector3 position)
        {
            // use supplied entity that creates a motor base 
            // with 2 active wheels and one caster
            Pioneer3DX robotBaseEntity = new Pioneer3DX(position);

            // specify mesh. 
            robotBaseEntity.State.Assets.Mesh = "Pioneer3dx.bos";
            // specify color if no mesh is specified. 
            robotBaseEntity.ChassisShape.State.DiffuseColor = new Vector4(0.8f, 0.25f, 0.25f, 1.0f);

            // the name below must match manifest
            robotBaseEntity.State.Name = "P3DXMotorBase";

            // Start simulated arcos motor service
            drive.Contract.CreateService(ConstructorPort,
                Microsoft.Robotics.Simulation.Partners.CreateEntityPartner(
                    "http://localhost/" + robotBaseEntity.State.Name)
            );
            return robotBaseEntity;
        }

        private LaserRangeFinderEntity CreateLaserRangeFinder()
        {
            // Create a Laser Range Finder Entity .
            // Place it 30cm above base CenterofMass. 
            LaserRangeFinderEntity laser = new LaserRangeFinderEntity(
                new Pose(new Vector3(0, 0.30f, 0)));

            laser.State.Name = "P3DXLaserRangeFinder";
            laser.LaserBox.State.DiffuseColor = new Vector4(0.25f, 0.25f, 0.8f, 1.0f);

            // Create LaserRangeFinder simulation service and specify
            // which entity it talks to
            simlrf.Contract.CreateService(
                ConstructorPort,
                Microsoft.Robotics.Simulation.Partners.CreateEntityPartner(
                "http://localhost/" + laser.State.Name));
            return laser;
        }

        private BumperArrayEntity CreateBumperArray()
        {
            // Create a bumper array entity with two bumpers
            BoxShape frontBumper = new BoxShape(
                new BoxShapeProperties("front",
                    0.001f,
                    new Pose(new Vector3(0, 0.05f, -0.25f)),
                    new Vector3(0.40f, 0.03f, 0.03f)
                )
            );
            frontBumper.State.DiffuseColor = new Vector4(0.1f, 0.1f, 0.1f, 1.0f);

            BoxShape rearBumper = new BoxShape(
                new BoxShapeProperties("rear",
                    0.001f,
                    new Pose(new Vector3(0, 0.05f, 0.25f)),
                    new Vector3(0.40f, 0.03f, 0.03f)
                )
            );
            rearBumper.State.DiffuseColor = new Vector4(0.1f, 0.1f, 0.1f, 1.0f);

            // The physics engine will issue contact notifications only
            // if we enable them per shape
            frontBumper.State.EnableContactNotifications = true;
            rearBumper.State.EnableContactNotifications = true;

            // put some force filtering so we only get notified for significant bumps
            //frontBumper.State.ContactFilter = new ContactNotificationFilter(1,1);
            //rearBumper.State.ContactFilter = new ContactNotificationFilter(1, 1);

            BumperArrayEntity
                bumperArray = new BumperArrayEntity(frontBumper, rearBumper);
            // entity name, must match manifest partner name
            bumperArray.State.Name = "P3DXBumpers";

            // start simulated bumper service

            bumper.Contract.CreateService(
                ConstructorPort,
                Microsoft.Robotics.Simulation.Partners.CreateEntityPartner(
                "http://localhost/" + bumperArray.State.Name));
            return bumperArray;
        }

        private CameraEntity CreateCamera()
        {
            // low resolution, wide Field of View
            CameraEntity cam = new CameraEntity(320, 240);
            cam.State.Name = "robocam";
            // just on top of the bot
            cam.State.Pose.Position = new Vector3(0.0f, 0.5f, 0.0f);
            // camera renders in an offline buffer at each frame
            // required for service
            cam.IsRealTimeCamera = true;

            // Start simulated webcam service
            simwebcam.Contract.CreateService(
                ConstructorPort,
                Microsoft.Robotics.Simulation.Partners.CreateEntityPartner(
                    "http://localhost/" + cam.State.Name)
            );

            return cam;
        }
        
        #endregion

    }

    #region Table Entity

    /// <summary>
    /// An entity for approximating a table. 
    /// </summary>
    [DataContract]
    public class TableEntity : MultiShapeEntity
    {
        /// <summary>
        /// Default constructor.
        /// </summary>
        public TableEntity() { }

        /// <summary>
        /// Custom constructor, programmatically builds physics primitive shapes to describe
        /// a particular table. 
        /// </summary>
        /// <param name="position"></param>
        public TableEntity(Vector3 position)
        {
            State.Pose.Position = position;
            State.Assets.Mesh = "table_01.obj";
            float tableHeight = 0.65f;
            float tableWidth = 1.05f;
            float tableDepth = 0.7f;
            float tableThinkness = 0.03f;
            float legThickness = 0.03f;
            float legOffset = 0.05f;

            // add a shape for the table surface
            BoxShape tableTop = new BoxShape(
                new BoxShapeProperties(30,
                new Pose(new Vector3(0, tableHeight, 0)),
                new Vector3(tableWidth, tableThinkness, tableDepth))
            );

            // add a shape for the left leg
            BoxShape tableLeftLeg = new BoxShape(
                new BoxShapeProperties(10, // mass in kg
                new Pose(
                    new Vector3(-tableWidth / 2 + legOffset, tableHeight / 2, 0)),
                new Vector3(legThickness, tableHeight + tableThinkness, tableDepth))
            );

            BoxShape tableRightLeg = new BoxShape(
                new BoxShapeProperties(10, // mass in kg
                new Pose(
                    new Vector3(tableWidth / 2 - legOffset, tableHeight / 2, 0)),
                new Vector3(legThickness, tableHeight + tableThinkness, tableDepth))
            );

            BoxShapes = new List<BoxShape>();
            BoxShapes.Add(tableTop);
            BoxShapes.Add(tableLeftLeg);
            BoxShapes.Add(tableRightLeg);
        }

        public override void Update(FrameUpdate update)
        {
            base.Update(update);
        }
    } 

    #endregion

}
